#!/usr/bin/env python

def p_add(pa, pb):
	(xa, ya) = pa
	(xb, yb) = pb
	return (xa + xb, ya + yb)

def p_sub(pa, pb):
	(xa, ya) = pa
	(xb, yb) = pb
	return (xa - xb, ya - yb)

def to_p_wh(r):
	(x, y, w, h) = r
	return ( (x, y), (w, h) )

def p_wh_to(p, wh):
	( (x, y), (w, h) ) = (p, wh)
	return (x, y, w, h)

def get_p(r):
	(x, y, w, h) = r
	return (x, y)

def get_wh(r):
	(x, y, w, h) = r
	return (w, h)

def to_p2(r):
	(x, y, w, h) = r
	p = (x, y)
	wh = (w, h)
	return ( p, p_add(p, wh) )

def p2_to(p, p2):
	(x, y) = p
	(w, h) = p_sub(p2, p)
	return (x, y, w, h)

def op_and(ra, rb):
	( (ax, ay), (ax2, ay2) ) = to_p2(ra)
	( (bx, by), (bx2, by2) ) = to_p2(rb)
	p = ( max(ax, bx), max(ay, by) )
	p2 = ( min(ax2, bx2), min(ay2, by2) )
	return p2_to(p, p2)

def is_area(r):
	(x, y, w, h) = r
	return w > 0 and h > 0

def in_rect(p, r):
	(px, py) = p
	( (x, y), (x2, y2) ) = to_p2(r)
	return x <= px and px < x2 and y <= py and py < y2

def r_in_rect(ri, ro):
	(p, p2) = to_p2(ri)
	p2 = p_sub( p2, (1, 1) )
	return in_rect(p, ro) and in_rect(p2, ro)

def p_in_rects(p, rects):
	for r in rects:
		if in_rect(p, r):
			return True
	return False

def frame_rects(r):
	(x, y, w, h) = r
	return [
		(x, y, w, 1),
		(x, y+h-1, w, 1),
		(x, y+1, 1, h-2),
		(x+w-1, y+1, 1, h-2),
	]

def rects_contain_rect(rects):
	rx = None
	for r in rects:
		( (x, y), (x2, y2) ) = to_p2(r)
		if rx == None:
			(rx, ry, rx2, ry2) = (x, y, x2, y2)
		else:
			(rx, ry, rx2, ry2) = ( min(rx, x), min(ry, y), max(rx2, x2), max(ry2, y2) )
	return p2_to( (rx, ry), (rx2, ry2) )

def cvt_fxy(fxy, r):
	(p, p2) = to_p2(r)
	return p2_to( fxy(*p), fxy(*p2) )

def gather(ra, rb): # return rects
	(x, y, w, h) = r = op_and(ra, rb)
	if w < 0 or h < 0: # ! don't use is_area()
		return [ ra, rb ]
	if r == ra:
		return [ rb ]
	if r == rb:
		return [ ra ]

	( (ax, ay), (ax2, ay2) ) = to_p2(ra)
	( (bx, by), (bx2, by2) ) = to_p2(rb)
	if ax == bx and ax2 == bx2:
		y = min(ay, by)
		y2 = max(ay2, by2)
		return [ p2_to( (ax, y), (ax2, y2) ) ]
	if ay == by and ay2 == by2:
		x = min(ax, bx)
		x2 = max(ax2, bx2)
		return [ p2_to( (x, ay), (x2, ay2) ) ]

	rects = []
	y = min(ay, by)
	h = max(ay2, by2) - y
	for dy in range(h):
		y_ = y + dy
		in_a = ( ay <= y_ and y_ < ay2 )
		in_b = ( by <= y_ and y_ < by2 )
		if in_a and in_b:
			(x, x2) = ( min(ax, bx), max(ax2, bx2) )
		elif in_a:
			(x, x2) = (ax, ax2)
		else:
			(x, x2) = (bx, bx2)

		if not rects or ( rects[-1][0] != x or rects[-1][0] + rects[-1][2] != x2 ):
			rects += [ (x, y_, x2-x, 1) ]
		else:
			(xx, yy, ww, hh) = rects[-1]
			rects[-1] = (xx, yy, ww, hh+1)
	return rects

def rect_sub(a, b): # a >= b
	((ax,ay),(ax2,ay2)) = to_p2(a)
	((bx,by),(bx2,by2)) = to_p2(b)
	ret = []
	if ay != by:
		ret += [ p2_to((ax,ay),(ax2,by)) ]
	if ay2 != by2:
		ret += [ p2_to((ax,by2),(ax2,ay2)) ]
	if ax2 != bx2:
		ret += [ p2_to((bx2,by),(ax2,by2)) ]
	if ax != bx:
		ret += [ p2_to((ax,by),(bx,by2)) ]
	return ret
		
def r_sub_rs(r, rs):
	lst = [r]
	for r_ in rs:
		lst_ = []
		for r in lst:
			a = op_and(r_, r)
			lst_ += rect_sub(r, a) if is_area(a) else [r]
		lst = lst_
		if lst == []:
			break
	return lst

def gather_rs2(rsa, rsb):
	lst = []
	for r in rsa:
		lst += r_sub_rs(r, rsb)

	for r in rsb:
		add = [r]
		for r_ in lst:
			g = gather(r, r_)
			if g != [r, r_]:
				lst.remove(r_)
				add = g
				break
		lst += add
	return lst
# EOF
